/*
 * Segment 4 Datenübernahme und Aufbereitung, SWE 4.5 Abfrage Puffer
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.dua.abfrpuffer;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;

import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.bitctrl.dua.DUAKonstanten;

/**
 * Enthaelt informationen ueber den Zustand eines Datenendgeraets.
 *
 * @author BitCtrl Systems GmbH, Bachraty
 *
 */
public class DatenEndgeraet {
	/**
	 * Konstruktor.
	 *
	 * @param soDe
	 *            SystemObjekt Datenendgeraet
	 * @param soAssoziiert
	 *            Assoziiertes System Objekt - FahrStreifen oder
	 *            AchsLastMessStelle
	 * @param station
	 *            Station zu der das DE gehoert
	 */
	public DatenEndgeraet(final SystemObject soDe, final SystemObject soAssoziiert, final Station station) {
		soDE = soDe;
		this.soAssoziiert = soAssoziiert;
		this.station = station;

		for (int i = 0; i < DatenEndgeraet.RING_PUFFER_SIZE; i++) {
			ringPufferStatusAenderungen[i] = AbfrPufferKonstanten.ZEIT_BEGINN;
		}
	}

	/**
	 * Konstruktor.
	 *
	 * @param soDe
	 *            SystemObjekt Datenendgeraet
	 * @param station
	 *            Station zu der das DE gehoert
	 */
	public DatenEndgeraet(final SystemObject soDe, final Station station) {
		soDE = soDe;
		this.station = station;

		for (int i = 0; i < DatenEndgeraet.RING_PUFFER_SIZE; i++) {
			ringPufferStatusAenderungen[i] = AbfrPufferKonstanten.ZEIT_BEGINN;
		}
	}

	private boolean nochNieDaten = true;
	/**
	 * Station, zu die das DE gehoert.
	 */
	private Station station;
	/**
	 * Das SystemObjekt das dem DE representiert.
	 */
	private SystemObject soDE;
	/**
	 * Das Fahrstreifen oder MessStelle die mit dem DE verbunden ist.
	 */
	private SystemObject soAssoziiert;
	/**
	 * Ist Archivierung eingestellt.
	 */
	private boolean archiviert;
	/**
	 * Ob man aus dem archiv den ZeitStempel aktualisiert hat.
	 */
	private boolean archivAbgefragt;
	/**
	 * De ist verbunden.
	 */
	private boolean online;
	/**
	 * Das DE antwortet mit einer Negative Quittung.
	 */
	private boolean antwortetNegQuitt;
	/**
	 * Intervall der Abrufe.
	 */
	private long erzeugungsIntervall;
	/**
	 * Anzahl der Erwarteten DatenSaetzen.
	 */
	private long anzahlErwarteteDatenSaetze;
	/**
	 * ZeitStempell des letzten Datensatzes.
	 */
	private long letzterDatensatzZeitStempel = AbfrPufferKonstanten.ZEIT_BEGINN;
	/**
	 * Groesse des Ring Puffers.
	 */
	static final int RING_PUFFER_SIZE = 10;
	/**
	 * Der RingPuffer fuer die StatusAenderungen.
	 */
	private final long[] ringPufferStatusAenderungen = new long[DatenEndgeraet.RING_PUFFER_SIZE];
	/**
	 * Index im Ringpuffer.
	 */
	private int ringPufferPointer;

	/**
	 * Erfragt, ob das DatenEndgeraet parametriert ist.
	 *
	 * @return true, wenn parametriert
	 */
	boolean istParametriert() {
		return archiviert && archivAbgefragt && (erzeugungsIntervall > 0);
	}

	/**
	 * Stellt dem Ihnalt des DEs in lesbare form Ist vom Zeitpunkt abhaengig
	 * wegen flackern.
	 */
	public String toString(final long zeitpunkt) {
		final String s1, s2, s3;
		if (soDE != null) {
			s1 = soDE.getPid();
		} else {
			s1 = "null";
		}
		if (soAssoziiert != null) {
			s2 = soAssoziiert.getPid();
		} else {
			s2 = "null";
		}

		final DateFormat datumFormat = new SimpleDateFormat(DUAKonstanten.ZEIT_FORMAT_GENAU_STR);
		s3 = datumFormat.format(new Date(letzterDatensatzZeitStempel));
		return String.format("%-40s - %-40s  @ %-10s INT %2dh ARCH %5b ONLINE %5b FLACKERT %5b\n", s1, s2, s3,
				erzeugungsIntervall / AbfrPufferKonstanten.STUNDE_IN_MS, archiviert, online, flackert(zeitpunkt));
	}

	/**
	 * Erfragt nach die Station.
	 *
	 * @return Station in der sich das DE befindet
	 */
	public Station getStation() {
		return station;
	}

	/**
	 * Stellt die zuhoerige Station fest.
	 *
	 * @param station
	 *            Station in der sich das DE befindet
	 */
	public void setStation(final Station station) {
		this.station = station;
	}

	/**
	 * Efragt nach dem SystemObjekt des DEs.
	 *
	 * @return Das passende SystemObjekt
	 */
	public SystemObject getSoDE() {
		return soDE;
	}

	/**
	 * Stellt das passende SystemObjekt des DEs fest.
	 *
	 * @param soDE
	 *            SystemObjekt DatenEndgeraet
	 */
	public void setSoDE(final SystemObject soDE) {
		this.soDE = soDE;
	}

	/**
	 * Erfragt nach dem assoziierten SystemObjekt (FS oder MS).
	 *
	 * @return the SystemObjekt Assoziiert
	 */
	public SystemObject getSoAssoziiert() {
		return soAssoziiert;
	}

	/**
	 * Setzt den assoziierten SystemObjekt ( Fahrstreifen oder
	 * Achslastmessstelle ).
	 *
	 * @param soAssoziiert
	 *            the SystemObjekt Assoziiert to set
	 */
	public void setSoAssoziiert(final SystemObject soAssoziiert) {
		this.soAssoziiert = soAssoziiert;
	}

	/**
	 * Erfragt, ob noch nie Daten fuer diese DE im Archiv standen bzw. empfangen
	 * wurden.
	 *
	 * @return <code>true</code>, wenn noch nie Daten fuer diese DE im Archiv
	 *         standen bzw. empfangen wurden.
	 */
	public boolean isNochNieDaten() {
		return nochNieDaten;
	}

	/**
	 * Erfragt den Flag ob die Daten von dem DE archiviert werden
	 *
	 * @return true, wenn archiviert
	 */
	public boolean isArchiviert() {
		return archiviert;
	}

	/**
	 * Setzt den Flag ob die Daten von dem DE archiviert werden
	 *
	 * @param archiviert
	 *            true, wenn archiviert
	 */
	public void setArchiviert(final boolean archiviert) {
		this.archiviert = archiviert;
		if (!archiviert) {
			archivAbgefragt = false;
		}
	}

	/**
	 * Erfragt ob das DE online ist
	 *
	 * @return true, wenn online
	 */
	public boolean isOnline() {
		return online;
	}

	/**
	 * Setzt ob das DE online ist
	 *
	 * @param online
	 *            true, wenn es online ist,
	 * @param zeitStempel
	 *            Zeitstempel der Aenderung
	 */
	public void setOnline(final boolean online, final long zeitStempel) {
		if (this.online != online) {
			ringPufferStatusAenderungen[ringPufferPointer] = zeitStempel;
			ringPufferPointer = ++ringPufferPointer % DatenEndgeraet.RING_PUFFER_SIZE;
		}
		this.online = online;
	}

	/**
	 * Erfragt ob das De flackert
	 *
	 * @param zeitStemepel
	 *            Zeitpunkt jetzt
	 * @return true, wenn die letzten 10 Statusaenderungen innerhalb von dem
	 *         minimalen Zeitintervall geschehen sind (lineare Funktion)
	 */
	public boolean flackert(final long zeitStemepel) {

		for (int i = 3; i < DatenEndgeraet.RING_PUFFER_SIZE; i++) {
			final int index = ((ringPufferPointer - i) + DatenEndgeraet.RING_PUFFER_SIZE)
					% DatenEndgeraet.RING_PUFFER_SIZE;
			long minDelta = AbfrPufferKonstanten.STATUS_AENDERUNGEN_MINIMUM_DELTA
					/ (DatenEndgeraet.RING_PUFFER_SIZE - 3);
			minDelta *= i;
			if ((zeitStemepel - ringPufferStatusAenderungen[index]) < minDelta) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Liefert den ErzeugungsIntervall
	 *
	 * @return ZeitIntervall
	 */
	public long getErzeugungsIntervall() {
		return erzeugungsIntervall;
	}

	/**
	 * Setzt den Erzeugungsintervall
	 *
	 * @param erzeugungsIntervall
	 *            Zeitintervall
	 */
	public void setErzeugungsIntervall(final long erzeugungsIntervall) {
		this.erzeugungsIntervall = erzeugungsIntervall;
	}

	/**
	 * Liefert den Anzahl der erwarteten DS
	 *
	 * @return Anzahl
	 */
	public long getAnzahlErwarteteDatenSaetze() {
		return anzahlErwarteteDatenSaetze;
	}

	/**
	 * Setzt den Anzahl der erwarteten DS
	 *
	 * @param anzahlErwarteteDatenSaetze
	 *            Anzahl
	 */
	public void setAnzahlErwarteteDatenSaetze(final long anzahlErwarteteDatenSaetze) {
		this.anzahlErwarteteDatenSaetze = anzahlErwarteteDatenSaetze;
	}

	/**
	 * Liefert den Zeitstemepl des letzten DS
	 *
	 * @return ZeitStempel
	 */
	public long getLetzterDatensatzZeitStempel() {
		return letzterDatensatzZeitStempel;
	}

	/**
	 * Decrementiert den Anzahl der Erwarteten Datensaetzen
	 *
	 * @return Die weitererwartete DatenSaetze
	 */
	public long decrementAnzahlErwarteteDatenSaetze() {
		if (anzahlErwarteteDatenSaetze > 0) {
			return --(anzahlErwarteteDatenSaetze);
		} else {
			return -1;
		}
	}

	/**
	 * Aktialisiert den letzten ZeitStempel
	 *
	 * @param letzterDatensatzZeitStempell
	 *            ZeitStempel
	 */
	public void updateLetzterDatensatzZeitStempel(final long letzterDatensatzZeitStempell) {
		nochNieDaten = false;
		if (letzterDatensatzZeitStempel < letzterDatensatzZeitStempell) {
			letzterDatensatzZeitStempel = letzterDatensatzZeitStempell;
		}
	}

	/**
	 * Liefert den Flag das der Archiv schon abgefragt wurde
	 *
	 * @return true, wenn der Archiv abgefragt wurde
	 */
	public boolean isArchivAbgefragt() {
		return archivAbgefragt;
	}

	/**
	 * Setzt den Flag das der Archiv schon abgefragt wurde
	 *
	 * @param archivAbgefragt
	 *            true, wenn der Archiv abgefragt wurde
	 */
	public void setArchivAbgefragt(final boolean archivAbgefragt) {
		this.archivAbgefragt = archivAbgefragt;
	}

	/**
	 * Erfragt ob das DE mit einer Negativen Quittung geantwortet hat
	 *
	 * @return negativeQuittung
	 */
	public boolean antwortetNegQuitt() {
		return antwortetNegQuitt;
	}

	/**
	 * Setzt den Flag ob das DE mit einer Negativen Quittung geantwortet hat
	 *
	 */
	public void setAntwortetNegQuitt(final boolean negQuitt) {
		antwortetNegQuitt = negQuitt;
	}
}